import os
import json
import shutil
from importlib import import_module
from crontab import CronTab


DEFAULT_CONF_FILE = os.path.join(os.path.dirname(__file__), 'CONFIG.JSON')
CONF_FILE = os.path.join(os.path.expanduser('~'), 'CONFIG.JSON')

ENABLE_W1_TEMPLATE="""
#!/bin/sh -e
#
# Start One Wire protocol
#
# You need sudo to run this script. 
# 
# This script should be called in /etc/rc.local to be run
# on each system startup. Alternatively, define a cronjob:
#
# >> sudo crontab -e
# >> @reboot /path/to/enable_w1.sh
#

{commands}
exit 0
"""


def get_serial_number():
    # dummy versions
    versions = dict(
        hardware='xxxxxxx',
        revision='0000000',
        serial='0000000000000000'
    )

    # open cpu info and read
    try:
        with open('/proc/cpuinfo', 'r') as info:
            for line in info:
                if line.startswith('Hardware'):
                    versions['hardware'] = line.split(':')[1].strip()
                elif line.startswith('Revision'):
                    versions['revision'] = line.split(':')[1].strip()
                elif line.startswith('Serial'):
                    versions['serial'] = line.split(':')[1].strip()

    except:
        versions = dict(
        hardware='ERRORxx',
        revision='ERROR00',
        serial='ERROR00000000000'
    )

    return versions


def parse_interval_to_seconds(s: str) -> int:
    s = s.lower()

    # Hours
    if 'hrs' in s:
        s = s.replace('hrs', 'h')
    if 'h' in s:
        s = s.replace('h', '')
        t = int(s)
        return t * 3600
    
    # Minutes
    if 'min' in s:
        s = s.replace('min', 'm')
    if 'm' in s:
        s = s.replace('m', '')
        t = int(s)
        return t * 60

    # Seconds
    if 'sec' in s:
        s = s.replace('sec', 's')
    if 's' in s:
        s = s.replace('s', '')
        return int(s)


def load_sensor(sensor_name: str):
    # try to load the module from raspi_logger.sensors
    try:
        return import_module('raspi_logger.sensors.%s' % sensor_name)
    except AttributeError:
        pass
    
    # if still here, load from globals
    if sensor_name in globals():
        return globals()['sensor_name']
    else:
        raise ValueError("A sensor of name '%s' could not be loaded." % sensor_name)


def config(**kwargs) -> dict:
    if not os.path.exists(CONF_FILE):
        shutil.copy(DEFAULT_CONF_FILE, CONF_FILE)
    
    # get the config
    with open(CONF_FILE, 'r') as f:
        conf = json.load(f)

    # check if we are in read or write mode
    if len(kwargs.keys()) == 0:
        return conf 
    else:
        conf.update(kwargs)

        # write
        with open(CONF_FILE, 'w') as f:
            json.dump(conf, f, indent=4)
        
        return conf


def reset_config():
    shutil.copy(DEFAULT_CONF_FILE, CONF_FILE)


def enable_w1(path=None, gpio=[4]):
    TEMPLATE = '# enable w1 on GPIO {pin}\ndtoverlay w1-gpio gpiopin={pin} pullup=0'
    try:
        if os.geteuid() != 0:
            raise AttributeError
    except AttributeError:
        print('You need root privileges on a UNIX OS to run this command.\nRun again like:\nsudo python3 -m raspi_logger enable_w1')
        return

    # get the script location
    if path is None:
        PATH = os.path.abspath(os.path.join(os.path.expanduser('~'), 'enable_w1.sh'))
    else:
        PATH = path

    # create the script
    cmds = '\n'.join([TEMPLATE.format(pin=pin) for pin in gpio])
    # copth the file over
    with open(PATH, 'w') as shellscript:
        shellscript.write(ENABLE_W1_TEMPLATE.format(commands=cmds))

    # make it executeable
    os.chmod(PATH, 0o755)
    
    # we run with sudo
    cron = CronTab(user='root')
    job = cron.new(command='%s' % PATH)
    job.every_reboot()
    cron.write()

    print('OneWire enabled. GPIO: %s' % (', '.join(gpio)))



import os
from datetime import datetime as dt

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column, BigInteger, Integer, String, Boolean, DateTime

from raspi_logger.util import config

Base = declarative_base()


class RawData(Base):
    __tablename__ = 'raw_data'

    id = Column(BigInteger().with_variant(Integer, 'sqlite'), primary_key=True)
    created=Column(DateTime, default=dt.now)
    raw_dump = Column(String, nullable=False)
    uploaded = Column(Boolean, default=False)


def connect(conf):
    engine = create_engine(path)
    SessionCls = sessionmaker(bind=engine)

    # create the correct connection path
    p = os.path.join(conf.get('loggerPath'), 'rawDataLogger.db')
    path = 'sqlite:///%s' % p
    
    # if the db file does not exist, a call of create_all is needed
    if not os.path.exists(p):
        Base.metadata.create_all(engine)

    return SessionCls()


def append_data(data, conf=None, path=None):
    # TODO: deprecate the path here
    # get config
    if conf is None:
        conf = config()

    # connect to db
    session = connect(conf=conf)
    
    # append the data
    if not isinstance(data, list):
        data = [data]
    
    try:
        session.add_all([RawData(raw_dump=record) for record in data])
        session.commit()
    except Exception as e:
        session.rollback()
        
        # TODO, some kind of logs would be necessary here
        print(str(e))
    
    return data
    

def read_data(limit=None, conf=None):
    # get config
    if conf is None:
        conf = config()

    # connect to db
    session = connect(conf=conf)

    query = session.query(RawData).order_by(RawData.created)
    if limit is not None:
        query = query.limit(limit)





import os
import json
from datetime import datetime as dt
from time import time, sleep
from crontab import CronTab

from .util import parse_interval_to_seconds, config, load_sensor
from raspi_logger.backends.sqlite_backend import append_data as append_sqlite
from raspi_logger.backends.json_backend import append_data as append_json
# TODO: build a load_backend util, just like used with sensors


def current_data(sensor='all', dry=False, **kwargs):
    # get the config
    conf = config()

    # get the sensor modules
    sensorBackends = conf.get('sensorBackends', {'ds18b20': {}})

    # data buffer
    data = []
    for sen, sensor_conf in sensors.items():
        if sen == 'all' or sen.lower() == sensor.lower():
            # load the sensor_module:
            mod = load_sensor(sensor_name=sen)
            data.extend(mod.read_sensor(conf=sensor_conf, **kwargs))
    
    # in dry runs, only return the data
    if dry:
        return data

    if len(data) == 0:
        print('[ERROR]: recieved no data. Need better logging. Sorry.')
        return []
    
    # check the registered backends
    for name, c in conf.get('loggerBackends', {}).items():
        # check if the backend is currently enabled
        if c.get('enabled', True):
            if name == 'json':
                append_json(data, conf, path)
            elif name == 'sqlite':
                append_json(data, conf, path)

    # return the data for reuse
    return data


def show_current_data(**kwargs):
    """
    Issue a new sensor reading, without saving.\n
    For reading new values **with** saving, use the 
    current_data function.
    """
    return current_data(dry=True, **kwargs)


def stream(interval=None, dry=False, **kwargs):
    # get the start time
    t1 = time()
    
    if interval is None:
        interval = config().get('loggerInterval', '1min')
    else:
        config(loggerInterval=interval)
    
    if isinstance(interval, str):
        interval = parse_interval_to_seconds(interval)
    
    data = current_data(dry=dry, **kwargs)

    # stringify
    outstr = json.dumps(data, indent=4)

    # print
    print(outstr)

    # sleep for the remaining time
    remain = interval - (time() - t1)
    if remain < 0:
        remain = 0
    sleep(remain)

    # call again
    stream(dry=dry, **kwargs)



import os
import json
import glob
from os.path import join as pjoin
from datetime import datetime as dt

from raspi_logger.util import config

def append_data(new_data, conf):
    """
    path is deprecated and will be removed. The file path has to be specified
    in the CONFIG file.
    """
    # get the path
    date = dt.now().date()
    fname = '%d_%d_%d_raw_log.json' % (date.year, date.month, date.day)

    # get path
    # TODO: depreacte conf as argument and read the config file
    path = conf.get('loggerPath', pjoin(os.path.expanduser('~'), 'logger'))

    if not os.path.exists(path):
        os.makedirs(path)
        config(loggerPath=path)
    
    # get the filename path
    path = pjoin(path, fname)

    # save data
    # get existing data
    try:
        with open(path, 'r') as f:
            old_data = json.load(f)
    except:
        old_data = []
    
    # append and save
    with open(path, 'w') as js:
        old_data.extend(new_data)
        json.dump(old_data, js, indent=4)


def read_data(limit=None, max_files=None):
    """
    
    """
    # read the config
    conf = config()

    if not 'loggerPath' in conf:
        print('[ERROR]: No data saved as JSON.')
        return []

    # get the file path
    path = conf['loggerPath']

    # get all files
    data_files = glob.glob(pjoin(path, '*.json'))
    
    # sort - latest one as **first** element
    data_files.sort(key=os.path.getctime, reverse=True)

    # limit the maximum number of files to be read
    if max_files is not None:
        end_at = max_files if max_files < len(data_files) else len(data_files) 
        data_files = data_files[:end_at]

    # read the data
    sorter = lambda c: dt.fromisoformat(c['tstamp']).timestamp()
    data = []
    for fname in data_files:
        # load file content
        with open(fname, 'r') as f:
            data = json.load(f) + data
        
        # check if we have enough
        if limit is None:
            continue
        elif len(data) > limit:
            # we have enough data
            data.sort(key=sorter, reverse=True)
            return data[:limit][::-1]
    
    # if we reach this line, there was less data than limit or no limit
    data.sort(key=sorter)
    return data
